@@ -245,8 +245,9 @@ class ToolsFrame(QFrame):
             for field in sorted(tools[name]):
                 keyname = '%s.%s' % (name, field)
                 value = tools[name][field]
-                if not value is '':
-                    ini.set(section, keyname, value)
+                # value may be bool if originating from hglib.tortoisehgtools()
+                if value != '':
+                    ini.set(section, keyname, str(value))
 
         # 2. Save the new guidefs
         for n, toollistwidget in enumerate(self.widgets):
@@ -645,7 +646,7 @@ class CustomConfigDialog(QDialog):
         index = 0
         if selecteditem:
             try:
-                index = items.index(selecteditem)
+                index = list(items).index(selecteditem)
             except ValueError:
                 pass
         combo = QComboBox()
@@ -727,6 +728,8 @@ class CustomToolConfigDialog(CustomConfigDialog):
             '- {ROOT}: The path to the current repository root.\n'
             '- {REV} / {REVID}: the selected revision number / '
             'hexadecimal revision id hash respectively.\n'
+            '- {SELECTEDFILES}: The list of files selected by the user on the '
+            'revision details file list.\n'
             '- {FILES}: The list of files touched by the selected revision.\n'
             '- {ALLFILES}: All the files tracked by Mercurial on the selected'
             ' revision.'))
@@ -871,3 +874,40 @@ class HookConfigDialog(CustomConfigDialog):
         if not command:
             return _('You must set a command to run.')
         return '' # No error
+
+
+def addCustomToolsSubmenu(menu,
+        ui, location, make, slot, label=_('Custom Tools')):
+    '''
+    Add a custom tools submenu to an existing menus
+
+    This can be used, for example, to add the custom tools submenu to the
+    different file context menus
+    '''
+    tools, toollist = hglib.tortoisehgtools(ui,
+        selectedlocation=location)
+    if not tools:
+        return
+    submenu = menu.addMenu(label)
+    submenu.triggered.connect(slot)
+    emptysubmenu = True
+    for name in toollist:
+        if name == '|':
+            submenu.addSeparator()
+            continue
+        info = tools.get(name, None)
+        if info is None:
+            continue
+        command = info.get('command', None)
+        if not command:
+            continue
+        label = info.get('label', name)
+        icon = info.get('icon', CustomToolConfigDialog._defaulticonname)
+        status = info.get('status', 'MAR!C?S')
+        a = make(label, None, frozenset(status),
+            icon=icon, inmenu=submenu)
+        if a is not None:
+            a.setData(name)
+            emptysubmenu = False
+    if emptysubmenu:
+        menu.removeAction(submenu.menuAction())
