@@ -25,16 +25,16 @@ class HgignoreDialog(QDialog):
 
     contextmenu = None
 
-    def __init__(self, repo, parent=None, *pats):
+    def __init__(self, repoagent, parent=None, *pats):
         'Initialize the Dialog'
         QDialog.__init__(self, parent)
         self.setWindowFlags(self.windowFlags()
             & ~Qt.WindowContextHelpButtonHint
             | Qt.WindowMaximizeButtonHint)
 
-        self.repo = repo
+        self._repoagent = repoagent
         self.pats = pats
-        self.setWindowTitle(_('Ignore filter - %s') % repo.displayname)
+        self.setWindowTitle(_('Ignore filter - %s') % repoagent.displayName())
         self.setWindowIcon(qtlib.geticon('ignore'))
 
         vbox = QVBoxLayout()
@@ -56,12 +56,13 @@ class HgignoreDialog(QDialog):
         hbox.addWidget(add, 0)
 
         # layer 2
+        repo = repoagent.rawRepo()
         hbox = QHBoxLayout()
         vbox.addLayout(hbox)
         ignorefiles = [repo.wjoin('.hgignore')]
         for name, value in repo.ui.configitems('ui'):
             if name == 'ignore' or name.startswith('ignore.'):
-                ignorefiles.append(os.path.expanduser(value))
+                ignorefiles.append(util.expandpath(value))
 
         filecombo = QComboBox()
         hbox.addWidget(filecombo)
@@ -123,6 +124,10 @@ class HgignoreDialog(QDialog):
         s = QSettings()
         self.restoreGeometry(s.value('hgignore/geom').toByteArray())
 
+    @property
+    def repo(self):
+        return self._repoagent.rawRepo()
+
     def eventFilter(self, obj, event):
         if obj != self.ignorelist:
             return False
@@ -248,8 +253,10 @@ class HgignoreDialog(QDialog):
 
         try:
             self.repo.thginvalidate()
+            self.repo.lfstatus = True
             wctx = self.repo[None]
             wctx.status(unknown=True)
+            self.repo.lfstatus = False
         except (EnvironmentError, error.RepoError), e:
             qtlib.WarningMsgBox(_('Unable to read repository status'),
                                 uni(str(e)), parent=self)
@@ -312,11 +319,3 @@ class HgignoreDialog(QDialog):
         s = QSettings()
         s.setValue('hgignore/geom', self.saveGeometry())
         QDialog.reject(self)
-
-def run(_ui, *pats, **opts):
-    from tortoisehg.util import paths
-    from tortoisehg.hgqt import thgrepo
-    repo = thgrepo.repository(_ui, path=paths.find_root())
-    if pats and pats[0].endswith('.hgignore'):
-        pats = []
-    return HgignoreDialog(repo, None, *pats)
