@@ -9,9 +9,9 @@ import os
 
 from mercurial import hg, util, error, context, merge, scmutil
 
-from tortoisehg.util import paths, hglib
+from tortoisehg.util import hglib
 from tortoisehg.hgqt.i18n import _
-from tortoisehg.hgqt import qtlib, wctxactions, visdiff, cmdui, fileview, thgrepo
+from tortoisehg.hgqt import qtlib, wctxactions, cmdui, filedata, fileview
 
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
@@ -47,21 +47,26 @@ class StatusWidget(QWidget):
     linkActivated = pyqtSignal(QString)
     showMessage = pyqtSignal(unicode)
     fileDisplayed = pyqtSignal(QString, QString)
+    grepRequested = pyqtSignal(unicode, dict)
+    runCustomCommandRequested = pyqtSignal(str, list)
 
-    def __init__(self, repo, pats, opts, parent=None, checkable=True,
-                 defcheck='MAR!S'):
+    def __init__(self, repoagent, pats, opts, parent=None, checkable=True,
+                 defcheck='commit'):
         QWidget.__init__(self, parent)
 
         self.opts = dict(modified=True, added=True, removed=True, deleted=True,
                          unknown=True, clean=False, ignored=False, subrepo=True)
         self.opts.update(opts)
-        self.repo = repo
+        self._repoagent = repoagent
         self.pats = pats
         self.checkable = checkable
         self.defcheck = defcheck
         self.pctx = None
         self.savechecks = True
         self.refthread = None
+        self.refreshWctxLater = QTimer(self, interval=10, singleShot=True)
+        self.refreshWctxLater.timeout.connect(self.refreshWctx)
+        self.partials = {}
 
         # determine the user configured status colors
         # (in the future, we could support full rich-text tags)
@@ -110,7 +115,7 @@ class StatusWidget(QWidget):
             val = statusTypes[s]
             if self.opts[val.name]:
                 st = st + s
-        self.statusfilter = StatusFilterButton(
+        self.statusfilter = StatusFilterActionGroup(
             statustext=st, types=StatusType.preferredOrder)
 
         if self.checkable:
@@ -118,7 +123,7 @@ class StatusWidget(QWidget):
             self.checkNoneTT = _('Uncheck all files')
             self.checkAllNoneBtn = QCheckBox()
             self.checkAllNoneBtn.setToolTip(self.checkAllTT)
-            self.checkAllNoneBtn.stateChanged.connect(self.checkAllNone)
+            self.checkAllNoneBtn.clicked.connect(self.checkAllNone)
 
         self.filelistToolbar = QToolBar(_('Status File List Toolbar'))
         self.filelistToolbar.setIconSize(QSize(16,16))
@@ -130,24 +135,23 @@ class StatusWidget(QWidget):
             self.filelistToolbar.addSeparator()
         self.filelistToolbar.addWidget(le)
         self.filelistToolbar.addSeparator()
-        self.filelistToolbar.addWidget(self.statusfilter)
+        self.filelistToolbar.addWidget(
+            createStatusFilterMenuButton(self.statusfilter, self))
         self.filelistToolbar.addSeparator()
         self.filelistToolbar.addWidget(self.refreshBtn)
-        self.actions = wctxactions.WctxActions(self.repo, self, checkable)
+        self.actions = wctxactions.WctxActions(self._repoagent, self, checkable)
+        self.actions.refreshNeeded.connect(self.refreshWctx)
+        self.actions.runCustomCommandRequested.connect(
+            self.runCustomCommandRequested)
         tv = WctxFileTree(self.repo, checkable=checkable)
         vbox.addLayout(hbox)
         vbox.addWidget(tv)
         split.addWidget(frame)
 
-        if self.pats:
-            def clearPattern():
-                self.pats = []
-                self.refreshWctx()
-                cpb.setVisible(False)
-                self.titleTextChanged.emit(self.getTitle())
-            cpb = QPushButton(_('Remove filter, show root'))
-            vbox.addWidget(cpb)
-            cpb.clicked.connect(clearPattern)
+        self.clearPatternBtn = QPushButton(_('Remove filter, show root'))
+        vbox.addWidget(self.clearPatternBtn)
+        self.clearPatternBtn.clicked.connect(self.clearPattern)
+        self.clearPatternBtn.setVisible(bool(self.pats))
 
         tv.setItemsExpandable(False)
         tv.setRootIsDecorated(False)
@@ -157,13 +161,7 @@ class StatusWidget(QWidget):
         tv.menuRequest.connect(self.onMenuRequest)
         le.textEdited.connect(self.setFilter)
 
-        def statusTypeTrigger(status):
-            status = str(status)
-            for s in statusTypes:
-                val = statusTypes[s]
-                self.opts[val.name] = s in status
-            self.refreshWctx()
-        self.statusfilter.statusChanged.connect(statusTypeTrigger)
+        self.statusfilter.statusChanged.connect(self.setStatusFilter)
 
         self.tv = tv
         self.le = le
@@ -180,18 +178,41 @@ class StatusWidget(QWidget):
         docf.setLayout(vbox)
         self.docf = docf
 
-        self.fileview = fileview.HgFileView(self.repo, self)
+        self.fileview = fileview.HgFileView(self._repoagent, self)
+        self.fileview.setModeVisible(fileview.AnnMode, False)
+        self.fileview.setShelveButtonVisible(True)
         self.fileview.showMessage.connect(self.showMessage)
         self.fileview.linkActivated.connect(self.linkActivated)
         self.fileview.fileDisplayed.connect(self.fileDisplayed)
         self.fileview.shelveToolExited.connect(self.refreshWctx)
-        self.fileview.setContext(self.repo[None])
+        self.fileview.newChunkList.connect(self.updatePartials)
+        self.fileview.chunkSelectionChanged.connect(self.chunkSelectionChanged)
+        self.fileview.grepRequested.connect(self.grepRequested)
         self.fileview.setMinimumSize(QSize(16, 16))
         vbox.addWidget(self.fileview, 1)
 
         self.split = split
         self.diffvbox = vbox
 
+    @property
+    def repo(self):
+        return self._repoagent.rawRepo()
+
+    def __get_defcheck(self):
+        if self._defcheck is None:
+            return 'MAR!S'
+        return self._defcheck
+
+    def __set_defcheck(self, newdefcheck):
+        if newdefcheck.lower() == 'amend':
+            newdefcheck = 'MAS'
+        elif newdefcheck.lower() in ('commit', 'qnew', 'qrefresh'):
+            newdefcheck = 'MAR!S'
+        self._defcheck = newdefcheck
+
+    defcheck = property(__get_defcheck, __set_defcheck)
+
+    @pyqtSlot()
     def checkAllNone(self):
         state = self.checkAllNoneBtn.checkState()
         if state == Qt.Checked:
@@ -205,10 +226,11 @@ class StatusWidget(QWidget):
             self.checkAllNoneBtn.setTristate(False)
 
     def getTitle(self):
+        name = self._repoagent.displayName()
         if self.pats:
-            return _('%s - status (selection filtered)') % self.repo.displayname
+            return _('%s - status (selection filtered)') % name
         else:
-            return _('%s - status') % self.repo.displayname
+            return _('%s - status') % name
 
     def loadSettings(self, qs, prefix):
         self.fileview.loadSettings(qs, prefix+'/fileview')
@@ -218,22 +240,72 @@ class StatusWidget(QWidget):
         self.fileview.saveSettings(qs, prefix+'/fileview')
         qs.setValue(prefix+'/state', self.split.saveState())
 
+    @pyqtSlot(QString, object)
+    def updatePartials(self, wfile, changes):
+        # remove files from the partials dictionary if they are not partial
+        # selections, in order to simplify refresh.
+        dels = []
+        for file, oldchanges in self.partials.iteritems():
+            assert file in self.tv.model().checked
+            if oldchanges.excludecount == 0:
+                self.tv.model().checked[file] = True
+                dels.append(file)
+            elif oldchanges.excludecount == len(oldchanges.hunks):
+                self.tv.model().checked[file] = False
+                dels.append(file)
+        for file in dels:
+            del self.partials[file]
+
+        wfile = hglib.fromunicode(wfile)
+        if changes is None:
+            if wfile in self.partials:
+                del self.partials[wfile]
+                self.chunkSelectionChanged()
+            return
+
+        if wfile in self.partials:
+            # merge selection state from old hunk list to new hunk list
+            oldhunks = self.partials[wfile].hunks
+            oldstates = dict([(c.fromline, c.excluded) for c in oldhunks])
+            for chunk in changes.hunks:
+                if chunk.fromline in oldstates:
+                    self.fileview.updateChunk(chunk, oldstates[chunk.fromline])
+        else:
+            # the file was not in the partials dictionary, so it is either
+            # checked (all changes enabled) or unchecked (all changes
+            # excluded).
+            if wfile not in self.getChecked():
+                for chunk in changes.hunks:
+                    self.fileview.updateChunk(chunk, True)
+        self.chunkSelectionChanged()
+        self.partials[wfile] = changes
+
+    @pyqtSlot()
+    def chunkSelectionChanged(self):
+        'checkbox state has changed via chunk selection'
+        # inform filelist view that the file selection state may have changed
+        model = self.tv.model()
+        if model:
+            model.layoutChanged.emit()
+            model.checkCountChanged.emit()
+
     @pyqtSlot(QPoint, object)
     def onMenuRequest(self, point, selected):
         menu = self.actions.makeMenu(selected)
-        if menu.exec_(point):
-            self.refreshWctx()
+        menu.exec_(point)
 
     def setPatchContext(self, pctx):
         if pctx != self.pctx:
+            # clear out the current checked state on next refreshWctx()
             self.savechecks = False
-        else:
-            self.savechecks = True
         self.pctx = pctx
 
-    def refreshWctx(self, synchronous=False):
+    @pyqtSlot()
+    def refreshWctx(self):
         if self.refthread:
+            self.refreshWctxLater.start()
             return
+        self.refreshWctxLater.stop()
         self.fileview.clearDisplay()
 
         # store selected paths or current path
@@ -255,13 +327,11 @@ class StatusWidget(QWidget):
         self.refreshBtn.setEnabled(False)
         self.progress.emit(*cmdui.startProgress(_('Refresh'), _('status')))
         self.refthread = StatusThread(self.repo, self.pctx, self.pats, self.opts)
-        if not synchronous:
-            self.refthread.finished.connect(self.reloadComplete)
-        self.refthread.showMessage.connect(self.showMessage)
+        self.refthread.finished.connect(self.reloadComplete)
+        self.refthread.showMessage.connect(self.reloadFailed)
         self.refthread.start()
-        if synchronous:
-            self.reloadComplete()
 
+    @pyqtSlot()
     def reloadComplete(self):
         self.refthread.wait()
         if self.checkable:
@@ -271,9 +341,23 @@ class StatusWidget(QWidget):
         if self.refthread.wctx is not None:
             self.updateModel(self.refthread.wctx, self.refthread.patchecked)
         self.refthread = None
+        if len(self.repo.parents()) > 1:
+            # nuke partial selections if wctx has a merge in-progress
+            self.partials = {}
+        match = self.le.text()
+        if match:
+            self.setFilter(match)
+
+    # better to handle error in reloadComplete in place of separate signal?
+    @pyqtSlot(QString)
+    def reloadFailed(self, msg):
+        qtlib.ErrorMsgBox(_('Failed to refresh'), msg, parent=self)
+
+    def isRefreshingWctx(self):
+        return bool(self.refthread)
 
     def canExit(self):
-        return not self.refthread
+        return not self.isRefreshingWctx()
 
     def updateModel(self, wctx, patchecked):
         self.tv.setSortingEnabled(False)
@@ -290,7 +374,9 @@ class StatusWidget(QWidget):
                        checked, self, checkable=self.checkable,
                        defcheck=self.defcheck)
         if self.checkable:
-            tm.checkToggled.connect(self.updateCheckCount)
+            tm.checkToggled.connect(self.checkToggled)
+            tm.checkCountChanged.connect(self.updateCheckCount)
+        self.savechecks = True
 
         oldtm = self.tv.model()
         self.tv.setModel(tm)
@@ -302,6 +388,11 @@ class StatusWidget(QWidget):
         if self.checkable:
             self.updateCheckCount()
 
+        # remove non-existent file from partials table because model changed
+        for file in self.partials.keys():
+            if file not in tm.checked:
+                del self.partials[file]
+
         for col in (COL_PATH, COL_STATUS, COL_MERGE_STATE):
             w = self.tv.sizeHintForColumn(col)
             self.tv.setColumnWidth(col, w)
@@ -343,10 +434,18 @@ class StatusWidget(QWidget):
             self.actions.allactions[0].trigger()
         elif status == 'S':
             self.linkActivated.emit(
-                u'subrepo:' + hglib.tounicode(self.repo.wjoin(path)))
+                u'repo:' + hglib.tounicode(self.repo.wjoin(path)))
         elif status in 'C?':
             qtlib.editfiles(self.repo, [path])
 
+    @pyqtSlot(str)
+    def setStatusFilter(self, status):
+        status = str(status)
+        for s in statusTypes:
+            val = statusTypes[s]
+            self.opts[val.name] = s in status
+        self.refreshWctx()
+
     @pyqtSlot(QString)
     def setFilter(self, match):
         model = self.tv.model()
@@ -354,7 +453,16 @@ class StatusWidget(QWidget):
             model.setFilter(match)
             self.tv.enablefilterpalette(bool(match))
 
+    @pyqtSlot()
+    def clearPattern(self):
+        self.pats = []
+        self.refreshWctx()
+        self.clearPatternBtn.setVisible(False)
+        self.titleTextChanged.emit(self.getTitle())
+
+    @pyqtSlot()
     def updateCheckCount(self):
+        'user has toggled one or more checkboxes, update counts and checkall'
         model = self.tv.model()
         if model:
             model.checkCount = len(self.getChecked())
@@ -364,8 +472,20 @@ class StatusWidget(QWidget):
                 state = Qt.Checked
             else:
                 state = Qt.PartiallyChecked
+            self.checkAllNoneBtn.setTristate(state == Qt.PartiallyChecked)
             self.checkAllNoneBtn.setCheckState(state)
 
+    @pyqtSlot(QString, bool)
+    def checkToggled(self, wfile, checked):
+        'user has toggled a checkbox, update partial chunk selection status'
+        wfile = hglib.fromunicode(wfile)
+        if wfile in self.partials:
+            if wfile == hglib.fromunicode(self.fileview.filePath()):
+                for chunk in self.partials[wfile].hunks:
+                    self.fileview.updateChunk(chunk, not checked)
+            else:
+                del self.partials[wfile]
+
     def checkAll(self):
         model = self.tv.model()
         if model:
@@ -381,13 +501,26 @@ class StatusWidget(QWidget):
         if model:
             checked = model.getChecked()
             if types is None:
-                return [f for f, v in checked.iteritems() if v]
+                files = []
+                for f, v in checked.iteritems():
+                    if f in self.partials:
+                        changes = self.partials[f]
+                        if changes.excludecount < len(changes.hunks):
+                            files.append(f)
+                    elif v:
+                        files.append(f)
+                return files
             else:
                 files = []
                 for row in model.getAllRows():
                     path, status, mst, upath, ext, sz = row
-                    if status in types and checked[path]:
-                        files.append(path)
+                    if status in types:
+                        if path in self.partials:
+                            changes = self.partials[path]
+                            if changes.excludecount < len(changes.hunks):
+                                files.append(path)
+                        elif checked[path]:
+                            files.append(path)
                 return files
         else:
             return []
@@ -410,9 +543,13 @@ class StatusWidget(QWidget):
             return
         path, status, mst, upath, ext, sz = row
         wfile = util.pconvert(path)
-        pctx = self.pctx and self.pctx.p1() or None
-        self.fileview.setContext(self.repo[None], pctx)
-        self.fileview.displayFile(wfile, status)
+        ctx = self.repo[None]
+        pctx = self.pctx and self.pctx.p1() or ctx.p1()
+        if status == 'S':
+            fd = filedata.createSubrepoData(ctx, pctx, wfile)
+        else:
+            fd = filedata.createFileData(ctx, pctx, wfile, status)
+        self.fileview.display(fd)
 
 
 class StatusThread(QThread):
@@ -559,11 +696,13 @@ class WctxFileTree(QTreeView):
         self._paletteswitcher.enablefilterpalette(enable)
 
 class WctxModel(QAbstractTableModel):
-    checkToggled = pyqtSignal()
+    checkCountChanged = pyqtSignal()
+    checkToggled = pyqtSignal(QString, bool)
 
     def __init__(self, wctx, ms, pctx, savechecks, opts, checked, parent,
                  checkable=True, defcheck='MAR!S'):
         QAbstractTableModel.__init__(self, parent)
+        self.partials = parent.partials
         self.checkCount = 0
         rows = []
         nchecked = {}
@@ -640,17 +779,20 @@ class WctxModel(QAbstractTableModel):
             return 0 # no child
         return len(self.rows)
 
-    def check(self, files, state=True):
+    def check(self, files, state):
         for f in files:
+            assert f in self.checked
             self.checked[f] = state
+            self.checkToggled.emit(f, state)
         self.layoutChanged.emit()
-        self.checkToggled.emit()
-        
+        self.checkCountChanged.emit()
+
     def checkAll(self, state):
         for data in self.rows:
             self.checked[data[0]] = state
+            self.checkToggled.emit(data[3], state)
         self.layoutChanged.emit()
-        self.checkToggled.emit()
+        self.checkCountChanged.emit()
 
     def columnCount(self, parent):
         if parent.isValid():
@@ -664,7 +806,14 @@ class WctxModel(QAbstractTableModel):
         path, status, mst, upath, ext, sz = self.rows[index.row()]
         if index.column() == COL_PATH:
             if role == Qt.CheckStateRole and self.checkable:
-                # also Qt.PartiallyChecked
+                if path in self.partials:
+                    changes = self.partials[path]
+                    if changes.excludecount == 0:
+                        return Qt.Checked
+                    elif changes.excludecount == len(changes.hunks):
+                        return Qt.Unchecked
+                    else:
+                        return Qt.PartiallyChecked
                 if self.checked[path]:
                     return Qt.Checked
                 else:
@@ -729,9 +878,18 @@ class WctxModel(QAbstractTableModel):
         for index in indexes:
             assert index.isValid()
             fname = self.rows[index.row()][COL_PATH]
-            self.checked[fname] = not self.checked[fname]
+            uname = self.rows[index.row()][COL_PATH_DISPLAY]
+            if fname in self.partials:
+                checked = 0
+                changes = self.partials[fname]
+                if changes.excludecount < len(changes.hunks):
+                    checked = 1
+            else:
+                checked = self.checked[fname]
+            self.checked[fname] = not checked
+            self.checkToggled.emit(uname, self.checked[fname])
         self.layoutChanged.emit()
-        self.checkToggled.emit()
+        self.checkCountChanged.emit()
 
     def sort(self, col, order):
         self.layoutAboutToBeChanged.emit()
@@ -813,6 +971,7 @@ class WctxModel(QAbstractTableModel):
         self.reset()
 
     def getChecked(self):
+        assert len(self.checked) == len(self.unfiltered)
         return self.checked.copy()
 
 def statusMessage(status, mst, upath):
@@ -844,8 +1003,9 @@ statusTypes = {
                      'status.removed', _('removed')),
     '?' : StatusType('unknown', 'shelve.ico', _('%s is not tracked (unknown)'),
                      'status.unknown', _('unknown')),
-    '!' : StatusType('deleted', 'menudelete.ico', _('%s is missing!'),
-                     'status.deleted', _('deleted')),
+    '!' : StatusType('deleted', 'menudelete.ico',
+                     _('%s is deleted by non-hg command, but still tracked'),
+                     'status.deleted', _('missing')),
     'I' : StatusType('ignored', 'ignore.ico', _('%s is ignored'),
                      'status.ignored', _('ignored')),
     'C' : StatusType('clean', '', _('%s is not modified (clean)'),
@@ -855,40 +1015,39 @@ statusTypes = {
 }
 
 
-class StatusFilterButton(QToolButton):
-    """Button with drop-down menu for status filter"""
+class StatusFilterActionGroup(QObject):
+    """Actions to switch status filter"""
+
     statusChanged = pyqtSignal(str)
 
-    def __init__(self, statustext, types=None, parent=None, **kwargs):
+    def __init__(self, statustext, types=None, parent=None):
+        super(StatusFilterActionGroup, self).__init__(parent)
         self._TYPES = 'MARSC'
         if types is not None:
             self._TYPES = types
-        #if 'text' not in kwargs:
-        #    kwargs['text'] = _('Status')
-        super(StatusFilterButton, self).__init__(
-            parent, popupMode=QToolButton.MenuButtonPopup,
-            icon=qtlib.geticon('hg-status'),
-            toolButtonStyle=Qt.ToolButtonTextBesideIcon, **kwargs)
 
-        self.clicked.connect(self.showMenu)
-        self._initactions(statustext)
-
-    def _initactions(self, text):
         self._actions = {}
-        menu = QMenu(self)
         for c in self._TYPES:
             st = statusTypes[c]
-            a = menu.addAction('%s %s' % (c, st.trname))
+            a = QAction('&%s %s' % (c, st.trname), self)
             a.setCheckable(True)
-            a.setChecked(c in text)
+            a.setChecked(c in statustext)
             a.toggled.connect(self._update)
             self._actions[c] = a
-        self.setMenu(menu)
 
     @pyqtSlot()
     def _update(self):
         self.statusChanged.emit(self.status())
 
+    def actions(self):
+        return [self._actions[c] for c in self._TYPES]
+
+    def isChecked(self, c):
+        return self._actions[c].isChecked()
+
+    def setChecked(self, c, checked):
+        self._actions[c].setChecked(checked)
+
     def status(self):
         """Return the text for status filter"""
         return ''.join(c for c in self._TYPES
@@ -901,17 +1060,32 @@ class StatusFilterButton(QToolButton):
         for c in self._TYPES:
             self._actions[c].setChecked(c in text)
 
+
+def createStatusFilterMenuButton(actiongroup, parent=None):
+    """Create button with drop-down menu for status filter"""
+    button = QToolButton(parent)
+    button.setIcon(qtlib.geticon('hg-status'))
+    button.setPopupMode(QToolButton.InstantPopup)
+    menu = QMenu(button)
+    menu.addActions(actiongroup.actions())
+    button.setMenu(menu)
+    return button
+
+
 class StatusDialog(QDialog):
     'Standalone status browser'
-    def __init__(self, repo, pats, opts, parent=None):
+    def __init__(self, repoagent, pats, opts, parent=None):
         QDialog.__init__(self, parent)
         self.setWindowIcon(qtlib.geticon('hg-status'))
+        self._repoagent = repoagent
+
         layout = QVBoxLayout()
         layout.setContentsMargins(0, 0, 0, 0)
         self.setLayout(layout)
         toplayout = QVBoxLayout()
-        toplayout.setContentsMargins(10, 10, 10, 0);
-        self.stwidget = StatusWidget(repo, pats, opts, self, checkable=False)
+        toplayout.setContentsMargins(10, 10, 10, 0)
+        self.stwidget = StatusWidget(repoagent, pats, opts, self,
+                                     checkable=False)
         toplayout.addWidget(self.stwidget, 1)
         layout.addLayout(toplayout)
 
@@ -922,6 +1096,9 @@ class StatusDialog(QDialog):
         self.stwidget.titleTextChanged.connect(self.setWindowTitle)
         self.stwidget.linkActivated.connect(self.linkActivated)
 
+        self._subdialogs = qtlib.DialogKeeper(StatusDialog._createSubDialog,
+                                              parent=self)
+
         self.setWindowTitle(self.stwidget.getTitle())
         self.setWindowFlags(Qt.Window)
         self.loadSettings()
@@ -931,17 +1108,13 @@ class StatusDialog(QDialog):
         QTimer.singleShot(0, self.stwidget.refreshWctx)
 
     def linkActivated(self, link):
-        link = hglib.fromunicode(link)
-        if link.startswith('subrepo:'):
-            from tortoisehg.hgqt.run import qtrun
-            from tortoisehg.hgqt import commit
-            qtrun(commit.run, self.stwidget.repo.ui, root=link[8:])
-        if link.startswith('shelve:'):
-            from tortoisehg.hgqt import shelve
-            dlg = shelve.ShelveDialog(self.stwidget.repo, self)
-            dlg.finished.connect(dlg.deleteLater)
-            dlg.exec_()
-            self.refresh()
+        link = unicode(link)
+        if link.startswith('repo:'):
+            self._subdialogs.open(link[len('repo:'):])
+
+    def _createSubDialog(self, uroot):
+        repoagent = self._repoagent.subRepoAgent(uroot)
+        return StatusDialog(repoagent, [], {}, parent=self)
 
     def loadSettings(self):
         s = QSettings()
@@ -964,11 +1137,3 @@ class StatusDialog(QDialog):
             return
         self.saveSettings()
         QDialog.reject(self)
-
-def run(ui, *pats, **opts):
-    from tortoisehg.util import paths
-    from tortoisehg.hgqt import thgrepo
-    repo = thgrepo.repository(ui, path=paths.find_root())
-    pats = hglib.canonpaths(pats)
-    os.chdir(repo.root)
-    return StatusDialog(repo, pats, opts)
