@@ -1,4 +1,4 @@
-# thread.py - A seprated thread to run Mercurial command
+# thread.py - A separate thread to run Mercurial commands
 #
 # Copyright 2009 Steve Borho <steve@borho.org>
 # Copyright 2010 Yuki KODAMA <endflow.net@gmail.com>
@@ -6,22 +6,18 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2, incorporated herein by reference.
 
-import os
 import Queue
-import time
-import urllib2
+import urllib2, urllib
 import socket
 import errno
 
 from PyQt4.QtCore import *
-from PyQt4.QtGui import *
 
-from mercurial import util, error
+from mercurial import util, error, subrepo
 from mercurial import ui as uimod
 
 from tortoisehg.util import thread2, hglib
 from tortoisehg.hgqt.i18n import _, localgettext
-from tortoisehg.hgqt import qtlib
 
 local = localgettext()
 
@@ -29,6 +25,35 @@ class DataWrapper(object):
     def __init__(self, data):
         self.data = data
 
+# reference as cmdcore.ProgressMessage instead of thread.ProgressMessage
+class ProgressMessage(tuple):
+    __slots__ = ()
+
+    def __new__(cls, topic, pos, item='', unit='', total=None):
+        return tuple.__new__(cls, (topic, pos, item, unit, total))
+
+    @property
+    def topic(self):
+        return self[0]  # unicode
+    @property
+    def pos(self):
+        return self[1]  # int or None
+    @property
+    def item(self):
+        return self[2]  # unicode
+    @property
+    def unit(self):
+        return self[3]  # unicode
+    @property
+    def total(self):
+        return self[4]  # int or None
+
+    def __repr__(self):
+        names = ('topic', 'pos', 'item', 'unit', 'total')
+        fields = ('%s=%r' % (n, v) for n, v in zip(names, self))
+        return '%s(%s)' % (self.__class__.__name__, ', '.join(fields))
+
+
 class UiSignal(QObject):
     writeSignal = pyqtSignal(QString, QString)
     progressSignal = pyqtSignal(QString, object, QString, QString, object)
@@ -59,9 +84,9 @@ class UiSignal(QObject):
         except EOFError:
             raise util.Abort(local._('response expected'))
 
-    def promptchoice(self, msg, choices, default):
+    def promptchoice(self, prompt, default):
         try:
-            r = self._waitresponse(msg, False, choices, default)
+            r = self._waitresponse(prompt, False, True, default)
             if r is None:
                 raise EOFError
             return r
@@ -118,63 +143,87 @@ class QtUi(uimod.ui):
         if not self.interactive(): return default
         return self.sig.prompt(msg, default)
 
-    def promptchoice(self, msg, choices, default=0):
+    def promptchoice(self, prompt, default=0):
         if not self.interactive(): return default
-        return self.sig.promptchoice(msg, choices, default)
+        return self.sig.promptchoice(prompt, default)
 
-    def getpass(self, prompt=_('password: '), default=None):
-        return self.sig.getpass(prompt, default)
+    def getpass(self, prompt=None, default=None):
+        return self.sig.getpass(prompt or _('password: '), default)
 
     def progress(self, topic, pos, item='', unit='', total=None):
         return self.sig.progress(topic, pos, item, unit, total)
 
 
+# QThread.finished is raised before running and finished attributes are set
+HAVE_QTBUG_30251 = 0x40800 <= QT_VERSION < 0x40805
+
 class CmdThread(QThread):
     """Run an Mercurial command in a background thread, implies output
     is being sent to a rendered text buffer interactively and requests
     for feedback from Mercurial can be handled by the user via dialog
     windows.
     """
+
+    serviceStateChanged = pyqtSignal(int)
+
     # (msg=str, label=str)
     outputReceived = pyqtSignal(QString, QString)
 
-    # (topic=str, pos=int, item=str, unit=str, total=int)
-    # pos and total are emitted as object, since they may be None
-    progressReceived = pyqtSignal(QString, object, QString, QString, object)
+    progressReceived = pyqtSignal(ProgressMessage)
 
     # result: -1 - command is incomplete, possibly exited with exception
     #          0 - command is finished successfully
     #          others - return code of command
     commandFinished = pyqtSignal(int)
 
-    def __init__(self, cmdline, display, parent=None):
+    def __init__(self, parent=None, cwd=None):
         super(CmdThread, self).__init__(parent)
 
-        self.cmdline = cmdline
-        self.display = display
+        self.cmdline = None
+        self._cwd = hglib.fromunicode(cwd)
+        self._uihandler = None
         self.ret = -1
-        self.abortbyuser = False
         self.responseq = Queue.Queue()
-        self.rawoutput = QStringList()
         self.topics = {}
         self.curstrs = QStringList()
         self.curlabel = None
-        self.timer = QTimer(self)
+        self.timer = QTimer(self, interval=100)
         self.timer.timeout.connect(self.flush)
-        self.timer.start(100)
+        self.started.connect(self.timer.start)
         self.finished.connect(self.thread_finished)
 
-    def abort(self):
-        if self.isRunning() and hasattr(self, 'thread_id'):
-            self.abortbyuser = True
+    def serviceState(self):
+        return 0  # NoService
+
+    def startService(self):
+        pass
+
+    def stopService(self):
+        pass
+
+    def startCommand(self, cmdline, uihandler):
+        assert not self.isCommandRunning() or HAVE_QTBUG_30251
+        self.cmdline = map(hglib.fromunicode, cmdline)
+        self._uihandler = uihandler
+        if self._cwd:
+            self.cmdline[0:0] = ['--cwd', self._cwd]
+        self.start()
+
+    def abortCommand(self):
+        if self.isCommandRunning() and hasattr(self, 'thread_id'):
             try:
                 thread2._async_raise(self.thread_id, KeyboardInterrupt)
             except ValueError:
                 pass
 
+    def isCommandRunning(self):
+        return self.isRunning()
+
+    @pyqtSlot()
     def thread_finished(self):
         self.timer.stop()
         self.flush()
+        self._uihandler = None
         self.commandFinished.emit(self.ret)
 
     def flush(self):
@@ -185,20 +234,20 @@ class CmdThread(QThread):
             keys = self.topics.keys()
             for topic in keys:
                 pos, item, unit, total = self.topics[topic]
-                self.progressReceived.emit(topic, pos, item, unit, total)
+                progress = ProgressMessage(unicode(topic), pos, unicode(item),
+                                           unicode(unit), total)
+                self.progressReceived.emit(progress)
                 if pos is None:
                     del self.topics[topic]
         else:
             # Close all progress bars
             for topic in self.topics:
-                self.progressReceived.emit(topic, None, '', '', None)
+                progress = ProgressMessage(unicode(topic), None)
+                self.progressReceived.emit(progress)
             self.topics = {}
 
     @pyqtSlot(QString, QString)
     def output_handler(self, msg, label):
-        if label != 'control':
-            self.rawoutput.append(msg)
-
         if label == self.curlabel:
             self.curstrs.append(msg)
         else:
@@ -215,34 +264,21 @@ class CmdThread(QThread):
     def interact_handler(self, wrapper):
         prompt, password, choices, default = wrapper.data
         prompt = hglib.tounicode(prompt)
+        uihandler = self._uihandler
         if choices:
-            dlg = QMessageBox(QMessageBox.Question,
-                        _('TortoiseHg Prompt'), prompt, parent=self.parent())
-            dlg.setWindowFlags(Qt.Sheet)
-            dlg.setWindowModality(Qt.WindowModal)
-            for index, choice in enumerate(choices):
-                button = dlg.addButton(hglib.tounicode(choice),
-                                       QMessageBox.ActionRole)
-                button.response = index
-                if index == default:
-                    dlg.setDefaultButton(button)
-            dlg.exec_()
-            button = dlg.clickedButton()
-            if button is 0:
+            parts = prompt.split('$$')
+            resps = [p[p.index('&') + 1].lower() for p in parts[1:]]
+            uihandler.setPrompt(prompt, uihandler.ChoiceInput, resps[default])
+            r = uihandler.getLineInput()
+            if r is None:
                 self.responseq.put(None)
             else:
-                self.responseq.put(button.response)
+                self.responseq.put(resps.index(r))
         else:
-            mode = password and QLineEdit.Password \
-                             or QLineEdit.Normal
-            text, ok = qtlib.getTextInput(self.parent(),
-                         _('TortoiseHg Prompt'),
-                         prompt.title(),
-                         mode=mode)
-            if ok:
-                text = hglib.fromunicode(text)
-            else:
-                text = None
+            mode = password and uihandler.PasswordInput \
+                             or uihandler.TextInput
+            uihandler.setPrompt(prompt, mode)
+            text = hglib.fromunicode(uihandler.getLineInput())
             self.responseq.put(text)
 
     def run(self):
@@ -254,20 +290,26 @@ class CmdThread(QThread):
         ui.sig.interactSignal.connect(self.interact_handler,
                 Qt.QueuedConnection)
 
-        if self.display:
-            cmd = '%% hg %s\n' % self.display
-        else:
-            cmd = '%% hg %s\n' % ' '.join(self.cmdline)
-        ui.write(cmd, label='control')
-
         try:
             # save thread id in order to terminate by KeyboardInterrupt
             self.thread_id = int(QThread.currentThreadId())
 
             for k, v in ui.configitems('defaults'):
                 ui.setconfig('defaults', k, '')
+            # disable worker because it only works in main thread:
+            #   signal.signal(signal.SIGINT, signal.SIG_IGN)
+            #   ValueError: signal only works in main thread
+            ui.setconfig('worker', 'numcpus', 1)
             self.ret = 255
             self.ret = hglib.dispatch(ui, self.cmdline) or 0
+        except subrepo.SubrepoAbort, e:
+            errormsg = str(e)
+            label = 'ui.error'
+            if e.subrepo:
+                label += ' subrepo=%s' % urllib.quote(e.subrepo)
+            ui.write_err(local._('abort: ') + errormsg + '\n', label=label)
+            if e.hint:
+                ui.write_err(local._('hint: ') + str(e.hint) + '\n', label=label)
         except util.Abort, e:
             ui.write_err(local._('abort: ') + str(e) + '\n')
             if e.hint:
@@ -312,6 +354,14 @@ class CmdThread(QThread):
         except error.LockUnavailable, inst:
             ui.warn(local._("abort: could not lock %s: %s\n") %
                 (inst.desc or inst.filename, inst.strerror))
+        except error.CommandError, inst:
+            if inst.args[0]:
+                ui.warn(local._("hg %s: %s\n") % (inst.args[0], inst.args[1]))
+            else:
+                ui.warn(local._("hg: %s\n") % inst.args[1])
+        except error.OutOfBandError, inst:
+            ui.warn(local._("abort: remote error:\n"))
+            ui.warn(''.join(inst.args))
         except error.RepoError, inst:
             ui.warn(local._("abort: %s!\n") % inst)
         except error.ResponseError, inst:
@@ -324,6 +374,11 @@ class CmdThread(QThread):
                 ui.warn("\n%r\n" % util.ellipsis(inst.args[1]))
         except error.RevlogError, inst:
             ui.warn(local._("abort: %s!\n") % inst)
+        except error.UnknownCommand, inst:
+            ui.warn(local._("hg: unknown command '%s'\n") % inst.args[0])
+        except error.InterventionRequired, inst:
+            ui.warn("%s\n" % inst)
+            self.ret = 1
         except socket.error, inst:
             ui.warn(local._("abort: %s!\n") % str(inst))
         except IOError, inst:
@@ -351,20 +406,7 @@ class CmdThread(QThread):
             else:
                 ui.warn(local._("abort: %s\n") % inst.strerror)
         except Exception, inst:
-            if 'THGDEBUG' in os.environ:
-                import traceback
-                traceback.print_exc()
             ui.write_err(str(inst) + '\n')
+            raise
         except KeyboardInterrupt:
             self.ret = -1
-
-        if self.ret == -1:
-            if self.abortbyuser:
-                msg = _('[command terminated by user %s]')
-            else:
-                msg = _('[command interrupted %s]')
-        elif self.ret:
-            msg = _('[command returned code %d %%s]') % int(self.ret)
-        else:
-            msg = _('[command completed successfully %s]')
-        ui.write(msg % time.asctime() + '\n', label='control')
